#!/usr/bin/env node
const shell = require('shelljs');
const semver = require('semver');
const updateNotifier = require('update-notifier');
const prompts = require('prompts');
const updateTagPackage = require('./package.json');
const updateNotifierConfig = {
    pkg: updateTagPackage,
    updateCheckInterval: 1000 * 60 * 60 * 24 // 1 day
};

// command line args
const [, , ...args] = process.argv;
const dryRun = args[0] === '--dry-run';
const versionOutput = args[0] === '--version';

(async function() {
    function clearScreen() {
        process.stdout.write('\033c');
    }

    function checkIfVersionOutput() {
        if (versionOutput) {
            console.log(`${updateTagPackage.name} ${updateTagPackage.version}`);
            process.exit();
        }
    }

    function disableShellOutput() {
        shell.config.silent = true;
    }

    function enableShellOutput() {
        shell.config.silent = false;
    }

    function quitIfGitIsMissing() {
        if (!shell.which('git')) {
            shell.echo('Error: This script requires git');
            shell.exit(1);
        }
    }

    function quitIfNoGitRepo() {
        if (shell.exec('git rev-parse --is-inside-work-tree').code !== 0) {
            shell.echo('Error: Not inside a git directory');
            shell.exit(1);
        }
    }

    function showMessageIfDryRun() {
        if (dryRun) {
            console.log('DRY RUN - NOTHING WILL BE PUSHED');
        }
    }

    function calculateTags() {
        const result = new Object();

        const tags = shell.exec('git tag').split('\n');
        const filteredTags = tags.filter(t => semver.valid(t) && semver.prerelease(t) === null);
        const sortedTags = filteredTags.sort((a,b) => { return semver.rcompare(a,b); });

        result.latest = sortedTags[0];

        if (result.latest === '') {
            console.log('Error: Repository has no tags yet. Please add initial tag by hand.');
            process.exit();
        }

        result.nextMajor = semver.inc(result.latest, 'major');
        result.nextMinor = semver.inc(result.latest, 'minor');
        result.nextPatch = semver.inc(result.latest, 'patch');

        const preMajor = tags.filter(t => semver.valid(t) && semver.major(t) === semver.major(result.nextMajor) && semver.prerelease(t) !== null);
        result.latestPreMajor = preMajor.sort((a,b) => { return semver.rcompare(a,b); })[0];

        const preMinor = tags.filter(t => semver.valid(t) && semver.minor(t) === semver.minor(result.nextMinor) && semver.prerelease(t) !== null);
        result.latestPreMinor = preMinor.sort((a,b) => { return semver.rcompare(a,b); })[0];

        result.nextPreMajor = semver.inc(result.latest, 'premajor', 'RC');
        if (result.latestPreMajor) {
            result.nextPreMajor = semver.inc(result.latestPreMajor, 'prerelease', 'RC');
        }

        result.nextPreMinor = semver.inc(result.latest, 'preminor', 'RC');
        if (result.latestPreMinor) {
            result.nextPreMinor = semver.inc(result.latestPreMinor, 'prerelease', 'RC');
        }
        
        return result;
    }


    async function askToPush(tag) {
        const question = {
            type: 'select',
            name: 'value',
            message: `Add tag [${tag}] and push to origin?`,
            choices: [{ title: `yes`, value: true }, { title: `no`, value: false }],
            initial: 1
        };

        let response = await prompts(question);
        if (response.value) {
            push(tag);
        }
    }

    function push(tag) {
        enableShellOutput();
        if (dryRun) {
            console.log('NOT PUSHING BECAUSE OF: --dry-run');
        } else {
            shell.exec(`git tag ${tag}`);
            shell.exec(`git push`);
            shell.exec(`git push --tags`);
        }
        process.exit();
    }

    async function showMenu(nextTags) {
        const major = nextTags.nextMajor;
        const minor = nextTags.nextMinor;
        const patch = nextTags.nextPatch;
        const preMajor = nextTags.nextPreMajor;
        const preMinor = nextTags.nextPreMinor;

        console.log(`  Latest release version \t [${nextTags.latest}]`);
        console.log(`  Latest major-rc version \t [${(nextTags.latestPreMajor) ? nextTags.latestPreMajor : 'n/a'}]`);
        console.log(`  Latest minor-rc version \t [${(nextTags.latestPreMinor) ? nextTags.latestPreMinor : 'n/a'}]`);

        if (preMajor == null || preMinor == null || patch == null || minor == null || major == null) {
            console.log(`  Error: Cannot evaluate version from newest tag. Is the tag a version tag? [${latestTag}]`);
            process.exit();
        }

        const question = {
            type: 'select',
            name: 'value',
            message: 'Select new tag:',
            choices: [
                { title: `major \t [${major}]`, value: major },
                { title: `minor \t [${minor}]`, value: minor },
                { title: `patch \t [${patch}]`, value: patch },
                { title: `major-rc \t [${preMajor}]`, value: preMajor },
                { title: `minor-rc \t [${preMinor}]`, value: preMinor },
                { title: `quit`, value: 'quit' }
            ],
            initial: 4
        };
        let response = await prompts(question);

        if (response.value === 'quit') {
            process.exit();
        }

        await askToPush(response.value);
    }

    clearScreen();
    updateNotifier(updateNotifierConfig).notify({ defer: false });
    checkIfVersionOutput();
    disableShellOutput();
    quitIfGitIsMissing();
    quitIfNoGitRepo();
    showMessageIfDryRun();
    await showMenu(calculateTags());
})();
